# create and customize my own env.
import torch
import numpy as np
import itertools
import matplotlib.pyplot as plt

import fire
import random
import multiprocessing
from loguru import logger
from pathlib import Path
import imageio

import warnings
warnings.filterwarnings("ignore", category = UserWarning)

from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor, wait, FIRST_COMPLETED

import gymnasium as gym
from gymnasium import spaces
from gymnasium.core import ObservationWrapper
from gymnasium.envs.registration import register

# MiniGrid imports
from minigrid.envs.babyai.goto import GoToSeq
from minigrid.core.world_object import Ball, Box, Key, Goal
from minigrid.core.constants import OBJECT_TO_IDX, COLOR_TO_IDX
from minigrid.envs.babyai.core.verifier import GoToInstr, ObjDesc, BeforeInstr, AndInstr
from minigrid.utils.baby_ai_bot import BabyAIBot

from memmap_replay_buffer import ReplayBuffer

# Allowed types and colors (no grey to distinguish from walls)
ALLOWED_TYPES = ["ball", "box", "key"]
ALLOWED_COLORS = ["red", "green", "blue", "purple", "yellow"]

# Object constructors
OBJ_CONSTRUCTORS = {
    "ball": Ball,
    "box": Box,
    "key": Key,
}

# Allowed types and colors (no grey to distinguish from walls)
ALLOWED_TYPES = ["ball", "box", "key"]
ALLOWED_COLORS = ["red", "green", "blue", "purple", "yellow"]

# Object constructors
OBJ_CONSTRUCTORS = {
    "ball": Ball,
    "box": Box,
    "key": Key,
}

class PinPad(GoToSeq):
    def __init__(self, num_objects, obj_seq, room_size=8, num_rows=1, num_cols=1, seed=None, **kwargs):
        """
        Args:
            num_objects: Number of distinct objects to generate (total pool across episodes).
            obj_seq: Specifies the objects to visit and in what order for this episode.
                         Every index must be in [0, num_objects - 1].
            room_size: Size of each room
            num_rows: Number of room rows
            num_cols: Number of room columns
            seed: Random seed for object generation
        """
        assert num_objects <= len(ALLOWED_TYPES) * len(ALLOWED_COLORS), \
            f"Max distinct objects is {len(ALLOWED_TYPES) * len(ALLOWED_COLORS)}"
        
        # Validate obj_indices
        assert len(obj_seq) >= 1, "obj_seq must have at least 1 element"
        assert all(0 <= idx < num_objects for idx in obj_seq), \
            f"All indices in obj_seq must be in [0, {num_objects - 1}]"
        
        self.num_objects = num_objects
        self.obj_seq = obj_seq
        
        # Generate all possible (type, color) combinations and sample
        all_combinations = list(itertools.product(ALLOWED_TYPES, ALLOWED_COLORS))
        
        if seed is not None:
            rng = random.Random(seed)
            self.object_specs = rng.sample(all_combinations, num_objects)
        else:
            # sample the first num_objects
            self.object_specs = all_combinations[:num_objects]
        
        # Create mapping: (type, color) -> object_index (0 to num_objects-1)
        self.obj_to_idx = {spec: idx for idx, spec in enumerate(self.object_specs)}
        
        # One-hot channels:
        # 0: empty cell, 1: agent, 2: wall cell, 3+: distinct objects
        self.num_channels = 3 + num_objects
        
        super().__init__(
            room_size=room_size, 
            num_rows=num_rows, 
            num_cols=num_cols, 
            num_dists=0,  # We handle objects ourselves
            **kwargs
        )

    def gen_mission(self):
        self.place_agent()
    
        # Place all objects in the room (not just the ones in the task sequence)
        for idx in range(self.num_objects):
            obj_type, color = self.object_specs[idx]
            obj = OBJ_CONSTRUCTORS[obj_type](color)
            self.place_in_room(0, 0, obj)
        
        # Build instruction based on obj_indices (which objects to visit and in what order)
        def make_goto(obj_idx):
            t, c = self.object_specs[obj_idx]
            return GoToInstr(ObjDesc(t, c))
        
        # number of objects to visit (i.e. subgoals)
        num_to_visit = len(self.obj_seq)
        
        if num_to_visit == 1:
            self.instrs = make_goto(self.obj_seq[0])
        
        elif num_to_visit == 2:
            # "go to A, then go to B"
            self.instrs = BeforeInstr(
                make_goto(self.obj_seq[0]), 
                make_goto(self.obj_seq[1])
            )
        
        elif num_to_visit == 3:
            # "go to A and go to B, then go to C"

            # AndInstr is a combination of instructions that can be executed in any order
            # BeforeInstr is a combination of instructions that must be executed in the specified order
            # BeforeInstr can only be used for 2 instructions, and MiniGrid doesn't support chaining BeforeInstr s we resort to AndInstr, which should be ok for the purpose of generating expert demonstrations for BC - we can just generate the subgoal GT lables post-hoc.
            self.instrs = BeforeInstr(
                AndInstr(make_goto(self.obj_seq[0]), make_goto(self.obj_seq[1])),
                make_goto(self.obj_seq[2])
            )
        
        elif num_to_visit == 4:
            # "go to A and go to B, then go to C and go to D"

            # AndInstr is a combination of instructions that can be executed in any order
            # BeforeInstr is a combination of instructions that must be executed in the specified order
            # BeforeInstr can only be used for 2 instructions, and MiniGrid doesn't support chaining BeforeInstr s we resort to AndInstr, which should be ok for the purpose of generating expert demonstrations for BC - we can just generate the subgoal GT lables post-hoc.
            self.instrs = BeforeInstr(
                AndInstr(make_goto(self.obj_seq[0]), make_goto(self.obj_seq[1])),
                AndInstr(make_goto(self.obj_seq[2]), make_goto(self.obj_seq[3]))
            )
        
        else:
            raise ValueError(f"Max 4 objects supported for sequenced missions, got {num_to_visit}")

class OneHotFullyObsWrapper(ObservationWrapper):
    """
    One-hot encoding of the fully observable grid.
    
    Channels:
        0: empty space (None cells)
        1: agent
        2: wall
        3+: distinct objects (indexed by environment's obj_to_idx)
    """
    
    def __init__(self, env):
        super().__init__(env)
        
        # Get num_channels from the unwrapped environment
        unwrapped = self.env.unwrapped
        assert hasattr(unwrapped, 'num_channels'), \
            "Environment must have 'num_channels' attribute (use PinPidEnv)"
        
        self.num_channels = unwrapped.num_channels
        self.obj_to_idx = unwrapped.obj_to_idx
        
        width = unwrapped.width
        height = unwrapped.height
        
        new_image_space = spaces.Box(
            low=0,
            high=1,
            shape=(width, height, self.num_channels),
            dtype="uint8",
        )
        
        self.observation_space = spaces.Dict({
            **self.observation_space.spaces,
            "one_hot": new_image_space,
        })
    
    def observation(self, obs):
        env = self.unwrapped
        width, height = env.width, env.height
        
        # Get full grid encoding: [object_type, color, state]
        full_grid = env.grid.encode()
        
        # Initialize one-hot: (width, height, num_channels)
        one_hot = np.zeros((width, height, self.num_channels), dtype=np.uint8)
        
        for x in range(width):
            for y in range(height):
                obj_type_id = full_grid[x, y, 0]
                color_id = full_grid[x, y, 1]
                
                if obj_type_id == OBJECT_TO_IDX["unseen"] or obj_type_id == 0:
                    # Empty / unseen -> channel 0
                    one_hot[x, y, 0] = 1
                elif obj_type_id == OBJECT_TO_IDX["wall"]:
                    # Wall -> channel 2
                    one_hot[x, y, 2] = 1
                elif obj_type_id in [OBJECT_TO_IDX["ball"], OBJECT_TO_IDX["box"], OBJECT_TO_IDX["key"]]:
                    # Distinct object -> look up by (type, color)
                    idx_to_type = {v: k for k, v in OBJECT_TO_IDX.items()}
                    idx_to_color = {v: k for k, v in COLOR_TO_IDX.items()}
                    
                    obj_type = idx_to_type[obj_type_id]
                    color = idx_to_color[color_id]
                    
                    if (obj_type, color) in self.obj_to_idx:
                        obj_idx = self.obj_to_idx[(obj_type, color)]
                        one_hot[x, y, 3 + obj_idx] = 1
                    else:
                        # Unknown object, treat as empty
                        one_hot[x, y, 0] = 1
                else:
                    # Other (door, goal, lava, etc.) -> treat as empty for now
                    one_hot[x, y, 0] = 1
        
        # Agent position -> channel 1
        agent_x, agent_y = env.agent_pos
        one_hot[agent_x, agent_y, :] = 0  # Clear any other channel
        one_hot[agent_x, agent_y, 1] = 1  # Set agent channel
        
        return {**obs, "one_hot": one_hot}

# helpers
def exists(val):
    return val is not None

def sample(prob):
    return random.random() < prob

def get_mission_length(env_id, seed):
    """
    Get the mission length for a given seed.
    Returns the length of the mission string.
    """
    env = gym.make(env_id, render_mode="rgb_array")
    env.reset(seed=seed)
    length = len(env.unwrapped.mission)
    env.close()
    return length



class BabyAIBotEpsilonGreedy:
    def __init__(self, env, random_action_prob = 0.):
        self.expert = BabyAIBot(env)
        self.random_action_prob = random_action_prob
        self.num_actions = env.action_space.n
        self.last_action = None

    def __call__(self, state):
        if sample(self.random_action_prob):
            action = torch.randint(0, self.num_actions, ()).item()
        else:
            action = self.expert.replan(self.last_action)

        self.last_action = action
        return action

def collect_single_episode(env_id, obj_seq, seed, num_steps, random_action_prob, state_shape, visualize=False):
    """
    Collect a single episode of demonstrations.
    
    Args:
        env_id: registered environment ID
        obj_seq: sequence of object indices to visit
        seed: random seed for environment reset
        num_steps: max steps per episode
        random_action_prob: probability of taking random action
        state_shape: shape of state observations
    
    Returns:
        tuple of (episode_state, episode_action, success, episode_length)
    """

    env = gym.make(env_id, render_mode="rgb_array", obj_seq=obj_seq)
    env = OneHotFullyObsWrapper(env)

    try:
        state_obs, _ = env.reset(seed=seed)
        
        episode_state = np.zeros((num_steps, *state_shape), dtype=np.float32)
        episode_action = np.zeros(num_steps, dtype=np.float32)
        episode_label = -np.ones(num_steps, dtype=np.float32)     # these are the per-step subgoal labels (i.e. object indices)

        expert = BabyAIBotEpsilonGreedy(env.unwrapped, random_action_prob=random_action_prob)

        if visualize:
            frames = []
        else:
            frames = None

        for _step in range(num_steps):
            try:
                action = expert(state_obs)

                # extract the intended subgoal
                if hasattr(expert.expert, 'stack') and expert.expert.stack:
                    # each entry in self.expert.stack is a SubGoal instance
                    # current subgoal is the last entry in the stack
                    subgoal = expert.expert.stack[-1]
                    
                    if subgoal.reason == "Explore":
                        obj_idx = -1   # -1 indicates explore
                    else:
                        assert subgoal.reason is None
                        assert isinstance(subgoal.datum, ObjDesc)
                        obj_idx = env.unwrapped.obj_to_idx[(subgoal.datum.type, subgoal.datum.color)]
            except Exception as e:
                print("Error: ", e)
                env.close()
                return None, None, None, None, False, 0

            episode_label[_step] = obj_idx
            episode_state[_step] = state_obs["one_hot"]
            episode_action[_step] = action

            state_obs, reward, terminated, truncated, info = env.step(action)

            if visualize: 
                frames.append(env.render())

            if terminated:
                env.close()
                return frames, episode_label, episode_state, episode_action, True, _step + 1

        env.close()
        return frames, episode_label, episode_state, episode_action, False, num_steps
    except Exception as e:
        print("Error: ", e)
        env.close()
        return None, None, None, None, False, 0

def collect_demonstrations(
    env_id = "PinPad-v0",
    num_objects = 4,            # number of distinct objects (across episodes)
    room_size = 8,              # size of each room (width and height)
    num_rows = 1,               # number of rows of rooms
    num_cols = 1,               # number of columns of rooms    
    obj_seqs = [[0,1,2]],     # list of object sequences to visit
    seeds = None,               # list of seeds (if None, uses range(num_seeds))
    num_seeds = 100,            # number of seeds (ignored if seeds is provided)
    num_steps = 100,            # max steps per episode
    random_action_prob = 0.05,
    num_workers = None,
    output_dir = "pinpad_demonstrations",
    visualize = False,
):
    """
    Collect demonstrations using BabyAI Bot for the PinPad environment.
    
    For each (obj_seq, seed) pair, collects one episode.
    Total episodes = len(obj_seqs) * len(seeds)
    
    Parallelized using ProcessPoolExecutor.
    """

    # Register the environment
    if env_id in gym.envs.registry:
        del gym.envs.registry[env_id]
    
    register(
        id=env_id,
        entry_point=PinPad,
        kwargs={
            "num_objects": num_objects, 
            "obj_seq": obj_seqs[0],  # placeholder, will be overridden at gym.make
            "room_size": room_size, 
            "num_rows": num_rows, 
            "num_cols": num_cols
        },
    )

    # Determine state shape from environment
    temp_env = gym.make(env_id, obj_seq=obj_seqs[0])
    temp_env = OneHotFullyObsWrapper(temp_env)
    state_shape = temp_env.observation_space['one_hot'].shape
    temp_env.close()

    logger.info(f"Detected state shape: {state_shape} for env {env_id}")

    # Display object specs
    # i.e. which object index corresponds to which (type, color)
    logger.info("Object specs:")
    for idx, spec in enumerate(temp_env.unwrapped.object_specs):
        logger.info(f"Object {idx}: {spec}")

    # Display each task
    logger.info(f"Tasks: {obj_seqs}")

    if not exists(num_workers):
        num_workers = multiprocessing.cpu_count()

    # Build list of seeds
    if seeds is None:
        seeds = list(range(num_seeds))
    
    # Create all (obj_seq, seed) pairs
    all_tasks = [(obj_seq, seed) for obj_seq in obj_seqs for seed in seeds]
    total_episodes = len(all_tasks)

    logger.info(f"Collecting {total_episodes} episodes ({len(obj_seqs)} Tasks x {len(seeds)} seeds)")

    successful = 0
    progressbar = tqdm(total=total_episodes)

    output_folder = Path(output_dir)

    fields = {
        'state': ('float', state_shape),
        'action': ('float', ()),
        'label': ('float', ()),          # subgoal labels (i.e. object indices)
    }

    buffer = ReplayBuffer(
        folder = output_folder,
        max_episodes = total_episodes,
        max_timesteps = num_steps,
        fields = fields,
        overwrite = True
    )

    # Parallel execution with bounded pending futures to avoid OOM
    max_pending = num_workers

    with ProcessPoolExecutor(max_workers=num_workers) as executor:
        task_iter = iter(all_tasks)
        futures = {}

        # Initial batch of submissions
        for _ in range(min(max_pending, total_episodes)):
            task = next(task_iter, None)
            if exists(task):
                obj_seq, seed = task
                future = executor.submit(
                    collect_single_episode, env_id, obj_seq, seed, 
                    num_steps, random_action_prob, state_shape, visualize
                )
                futures[future] = task

        # Process completed tasks and submit new ones
        while futures:
            done, _ = wait(futures, return_when=FIRST_COMPLETED)

            for future in done:
                del futures[future]
                frames, episode_label, episode_state, episode_action, success, episode_length = future.result()

                if success and exists(episode_state):
                    buffer.store_episode(
                        state = episode_state[:episode_length],
                        action = episode_action[:episode_length],
                        label = episode_label[:episode_length],
                    )

                    # if visualize, save frames as .gif
                    # assert frames is not None
                    assert frames is not None
                    if visualize:
                        frames_path = output_folder / f"episode_{successful}.gif"
                        imageio.mimsave(frames_path, frames, fps=3)
                    
                    successful += 1

                progressbar.update(1)
                progressbar.set_description(f"success rate = {successful}/{progressbar.n:.2f}")

                # Submit a new task to replace the completed one
                task = next(task_iter, None)
                if exists(task):
                    obj_seq, seed = task
                    new_future = executor.submit(
                        collect_single_episode, env_id, obj_seq, seed,
                        num_steps, random_action_prob, state_shape, visualize
                    )
                    futures[new_future] = task

    buffer.flush()
    progressbar.close()

    # Save metadata for reproducibility
    metadata = {
        'obj_seqs': obj_seqs,
        'seeds': seeds,
        'num_objects': num_objects,
        'room_size': room_size,
    }
    metadata_path = output_folder / "metadata.npy"
    np.save(metadata_path, metadata, allow_pickle=True)

    logger.info(f"Saved {successful} trajectories to {output_dir}")
    logger.info(f"Saved metadata to {metadata_path}")

if __name__ == "__main__":
    fire.Fire(collect_demonstrations)

    """
    # How to run this script in terminal:

    python gather_pinpad_trajs.py \
        --num_objects 4 \   
        --obj_seqs "[[0,1,2],[0,2,1],[2,3,1],[0,2,3]]" \     # tasks i.e. list of object sequences to visit
        --room_size 8 \
        --num_rows 1 \
        --num_cols 1 \
        --num_seeds 10 \
        --num_steps 100 \
        --output_dir pinpad_demonstrations \
        --visualize True
    """


import string
import torch
import numpy as np
from functools import lru_cache
from pathlib import Path
from shutil import rmtree

import gymnasium as gym
from gymnasium import spaces
from gymnasium.core import Wrapper

import minigrid
from minigrid.wrappers import FullyObsWrapper, SymbolicObsWrapper

from sentence_transformers import SentenceTransformer

# Mission space compatible with gymnasium shared memory (for AsyncVectorEnv)
BABYAI_MISSION_TEXT_MAX_LENGTH = 512
BABYAI_MISSION_CHARSET = " " + string.ascii_lowercase + string.ascii_uppercase + string.digits + "',.-"


class BabyAISharedMemoryWrapper(Wrapper):
    """
    Replaces the BabyAI mission observation space (BabyAIMissionSpace) with
    gymnasium.spaces.Text so that AsyncVectorEnv(..., shared_memory=True) works.
    Observation values are unchanged: mission remains a string.
    """

    def __init__(self, env, max_mission_length=BABYAI_MISSION_TEXT_MAX_LENGTH, charset=BABYAI_MISSION_CHARSET):
        super().__init__(env)
        if "mission" not in env.observation_space.spaces:
            return
        mission_space = env.observation_space.spaces["mission"]
        # Replace only if it's minigrid's MissionSpace (no create_shared_memory in gymnasium)
        if type(mission_space).__name__ in ("BabyAIMissionSpace", "MissionSpace"):
            self._mission_text_space = spaces.Text(max_length=max_mission_length, charset=charset)
            self.observation_space = spaces.Dict(
                {
                    **{k: v for k, v in env.observation_space.spaces.items() if k != "mission"},
                    "mission": self._mission_text_space,
                }
            )
        else:
            self._mission_text_space = None

    def _process_obs(self, obs):
        if self._mission_text_space is None:
            return obs
        # Mission stays as string; Text space accepts it for shared memory encode
        mission = obs.get("mission", "")
        if len(mission) > self._mission_text_space.max_length:
            mission = mission[: self._mission_text_space.max_length]
        obs = dict(obs)
        obs["mission"] = mission
        return obs

    def reset(self, *, seed=None, options=None):
        obs, info = self.env.reset(seed=seed, options=options)
        return self._process_obs(obs), info

    def step(self, action):
        obs, reward, terminated, truncated, info = self.env.step(action)
        return self._process_obs(obs), reward, terminated, truncated, info

# sbert

_sbert = None

def get_sbert(local_files_only: bool = False, device: str = "cpu"):
    global _sbert
    if _sbert is None:
        _sbert = SentenceTransformer('all-MiniLM-L6-v2', local_files_only=local_files_only, device=device)
    return _sbert

@lru_cache(maxsize = 1000)
def get_mission_embedding(mission: str, sbert: object = None):
    if not sbert: sbert = get_sbert()
    embedding = sbert.encode(mission, convert_to_tensor = True)
    return embedding

_mission_cache = {}

def get_missions_embeddings(missions: list[str], sbert: object = None, batch_size: int = 64):
    if not sbert: sbert = get_sbert()
    
    # identify uncached missions
    
    unique_missions = list(set(missions))
    uncached_missions = [m for m in unique_missions if m not in _mission_cache]
    
    if len(uncached_missions) > 0:
        embeddings = sbert.encode(uncached_missions, convert_to_tensor=True, batch_size=batch_size, show_progress_bar=True)
        
        for mission, embedding in zip(uncached_missions, embeddings):
            _mission_cache[mission] = embedding

    # return stacked embeddings

    return torch.stack([_mission_cache[m] for m in missions])

# functions

def divisible_by(num, den):
    return (num % den) == 0

def transform_to_symbolic(images):
    # images: (..., W, H, 3) - assume last channel is (type, color, state)
    # as returned by FullyObsWrapper (which includes the agent)
    
    # We want (..., W, H, 3) where:
    # ch 0: x coord
    # ch 1: y coord
    # ch 2: object type
    
    *batch_dims, w, h, c = images.shape
    
    grid_x, grid_y = np.mgrid[:w, :h]
    
    # broadcast grid_x, grid_y to batch dims
    
    for _ in range(len(batch_dims)):
        grid_x = np.expand_dims(grid_x, axis = 0)
        grid_y = np.expand_dims(grid_y, axis = 0)

    grid_x = np.broadcast_to(grid_x, (*batch_dims, w, h))
    grid_y = np.broadcast_to(grid_y, (*batch_dims, w, h))

    images = images.astype(np.int64)

    # Output array
    symbolic = np.zeros(images.shape, dtype = np.int64)
    symbolic[..., 0] = grid_x
    symbolic[..., 1] = grid_y
    symbolic[..., 2] = np.where(images[..., 0] == 1, -1, images[..., 0]) # map empty (1) to -1
    
    return symbolic

# env creation

def create_env(
    env_id,
    render_mode = 'rgb_array',
    video_folder = None,
    render_every_eps = 1000,
    use_symbolic = True
):
    # register minigrid environments if needed
    # minigrid.register_minigrid_envs()

    # environment
    env = gym.make(env_id, render_mode = render_mode)
    env = FullyObsWrapper(env)

    if use_symbolic:
        env = SymbolicObsWrapper(env)

    env = BabyAISharedMemoryWrapper(env)

    if video_folder is not None:
        video_folder = Path(video_folder)
        rmtree(video_folder, ignore_errors = True)

        env = gym.wrappers.RecordVideo(
            env = env,
            video_folder = str(video_folder),
            name_prefix = 'babyai',
            episode_trigger = lambda eps_num: divisible_by(eps_num, render_every_eps),
            disable_logger = True
        )

    return env



from __future__ import annotations

import torch
from torch import nn, Tensor
from torch.nn import Module, ModuleList
from einops import rearrange, reduce
from einops.layers.torch import Rearrange, Reduce

from metacontroller.metacontroller import Transformer, Encoder

from torch_einops_utils import pack_with_inverse
from torch_einops_utils.save_load import save_load

# resnet components

def exists(v):
    return v is not None

class BasicBlock(Module):
    expansion = 1

    def __init__(
        self,
        dim,
        dim_out,
        stride = 1,
        downsample: Module | None = None
    ):
        super().__init__()
        self.conv1 = nn.Conv2d(dim, dim_out, 3, stride = stride, padding = 1, bias = False)
        self.bn1 = nn.BatchNorm2d(dim_out)
        self.relu = nn.ReLU(inplace = True)
        self.conv2 = nn.Conv2d(dim_out, dim_out, 3, padding = 1, bias = False)
        self.bn2 = nn.BatchNorm2d(dim_out)
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if exists(self.downsample):
            identity = self.downsample(x)

        out += identity
        return self.relu(out)

class Bottleneck(Module):
    expansion = 4

    def __init__(
        self,
        dim,
        dim_out,
        stride = 1,
        downsample: Module | None = None
    ):
        super().__init__()
        width = dim_out # simple resnet shortcut
        self.conv1 = nn.Conv2d(dim, width, 1, bias = False)
        self.bn1 = nn.BatchNorm2d(width)
        self.conv2 = nn.Conv2d(width, width, 3, stride = stride, padding = 1, bias = False)
        self.bn2 = nn.BatchNorm2d(width)
        self.conv3 = nn.Conv2d(width, dim_out * self.expansion, 1, bias = False)
        self.bn3 = nn.BatchNorm2d(dim_out * self.expansion)
        self.relu = nn.ReLU(inplace = True)
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if exists(self.downsample):
            identity = self.downsample(x)

        out += identity
        return self.relu(out)

class ResNet(Module):
    def __init__(
        self,
        block: type[BasicBlock | Bottleneck],
        layers: list[int],
        num_classes = 1000,
        channels = 3
    ):
        super().__init__()
        self.inplanes = 64

        self.conv1 = nn.Conv2d(channels, 64, kernel_size = 7, stride = 2, padding = 3, bias = False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace = True)
        self.maxpool = nn.MaxPool2d(kernel_size = 3, stride = 2, padding = 1)

        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride = 2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride = 2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride = 2)

        self.feature_dim = 512 * block.expansion

        self.fc = nn.Linear(self.feature_dim, num_classes)

    def _make_layer(
        self,
        block: type[BasicBlock | Bottleneck],
        planes: int,
        blocks: int,
        stride: int = 1
    ) -> nn.Sequential:
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion, 1, stride = stride, bias = False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x, attn: Encoder | None = None):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        if exists(attn):
            x = rearrange(x, 'b c h w -> b (h w) c')
            x = attn(x)
            x = reduce(x, 'b ... c ->  b c', 'mean')
        else:
            x = reduce(x, 'b c ... -> b c', 'mean')

        x = self.fc(x)
        return x

# resnet factory

def resnet18(num_classes = 1000):
    return ResNet(BasicBlock, [2, 2, 2, 2], num_classes)

def resnet34(num_classes = 1000):
    return ResNet(BasicBlock, [3, 4, 6, 3], num_classes)

def resnet50(num_classes = 1000):
    return ResNet(Bottleneck, [3, 4, 6, 3], num_classes)

# transformer with resnet

@save_load()
class TransformerWithResnet(Transformer):
    def __init__(
        self,
        *args,
        resnet_type = 'resnet18',
        is_channel_last = True,
        encoder_kwargs: dict | None = None,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.is_channel_last = is_channel_last

        resnet_klass = resnet18
        if resnet_type == 'resnet34':
            resnet_klass = resnet34
        elif resnet_type == 'resnet50':
            resnet_klass = resnet50

        self.resnet_dim = kwargs['state_embed_readout']['num_continuous']
        self.visual_encoder = resnet_klass(num_classes = self.resnet_dim)

        self.attn = None
        if exists(encoder_kwargs):
            assert 'dim' not in encoder_kwargs
            encoder_kwargs['dim'] = self.visual_encoder.feature_dim

            self.attn = Encoder(**encoder_kwargs)

    def visual_encode(self, x):
        if self.is_channel_last:
            x = rearrange(x, '... h w c -> ... c h w')

        x, inverse = pack_with_inverse(x, '* c h w')

        h = self.visual_encoder(x, attn = self.attn)

        return inverse(h, '* d')



from __future__ import annotations
from contextlib import nullcontext

from functools import partial
from collections import namedtuple
from loguru import logger

import torch
from torch import nn, cat, stack, tensor, Tensor
from torch.nn import Module, GRU, Linear, Identity
import torch.nn.functional as F

# einops

import einx
from einops import einsum, rearrange, repeat, reduce
from einops.layers.torch import Rearrange

# external modules

from x_transformers import Encoder, Decoder
from x_transformers.x_transformers import PolarEmbedding

from x_mlps_pytorch import Feedforwards
from x_evolution import EvoStrategy

from discrete_continuous_embed_readout import Embed, Readout, EmbedAndReadout

from assoc_scan import AssocScan

from torch_einops_utils import maybe, pad_at_dim, lens_to_mask, masked_mean, align_dims_left
from torch_einops_utils.device import module_device, move_inputs_to_module_device
from torch_einops_utils.save_load import save_load

# constants

LinearNoBias = partial(Linear, bias = False)

GRU = partial(GRU, batch_first = True)

# helper functions

def exists(v):
    return v is not None

def identity(t):
    return t

def default(*args):
    for arg in args:
        if exists(arg):
            return arg
    return None

# tensor helpers

def straight_through(src, tgt):
    return tgt + src - src.detach()

# action proposer wrapper
# normalizes any action proposer to a standard interface for MetaController

class ActionProposerWrapper(Module):
    def __init__(
        self,
        module: Module,
        cache_key = 'cache',
        return_cache_key = 'return_hiddens'
    ):
        super().__init__()
        self.module = module
        self.cache_key = cache_key
        self.return_cache_key = return_cache_key

    def forward(
        self,
        x,
        cache = None
    ):
        kwargs = {self.return_cache_key: True}

        if exists(cache):
            kwargs[self.cache_key] = cache

        out = self.module(x, **kwargs)

        return out if isinstance(out, tuple) else (out, None)

# losses

BehavioralCloningLosses = namedtuple('BehavioralCloningLosses', (
    'state',
    'action'
))

DiscoveryLosses = namedtuple('DiscoveryLosses', (
    'action_recon',
    'kl',
    'switch'
))

# meta controller

MetaControllerOutput = namedtuple('MetaControllerOutput', (
    'prev_hiddens',
    'input_residual_stream',
    'action_dist',
    'actions',
    'switch_beta',
    'kl_loss',
    'switch_loss'
))

GRPOOutput = namedtuple('GRPOOutput', (
    'state',
    'action',
    'log_prob',
    'switch_beta'
))

def z_score(t, dim = None, eps = 1e-8):
    kwargs = dict(dim = dim, keepdim = True) if exists(dim) else dict()
    return (t - t.mean(**kwargs)) / (t.std(**kwargs) + eps)

@move_inputs_to_module_device
def policy_loss(
    meta_controller,
    state,
    old_log_probs,
    actions,
    advantages,
    mask,
    episode_lens = None,
    eps_clip = 0.2
):
    # if switch betas sent as mask without converting to bool, just take care of it

    if mask.dtype == torch.float:
        mask = mask > 0.5

    # get new log probs

    action_dist = meta_controller.get_action_dist_for_internal_rl(state)
    new_log_probs = meta_controller.log_prob(action_dist, actions)

    # calculate ratio

    ratio = (new_log_probs - old_log_probs).exp()

    # align ratio and advantages

    ratio, advantages = align_dims_left((ratio, advantages))

    # ppo surrogate loss

    surr1 = ratio * advantages
    surr2 = ratio.clamp(1 - eps_clip, 1 + eps_clip) * advantages

    losses = -torch.min(surr1, surr2)

    # masking

    if exists(episode_lens):
        mask, episode_mask = align_dims_left((mask, lens_to_mask(episode_lens, losses.shape[1])))
        mask = mask & episode_mask

    return masked_mean(losses, mask)

def extract_grpo_data(meta_controller, transformer_output):
    meta_output = transformer_output.prev_hiddens.meta_controller

    state = meta_output.input_residual_stream
    action = meta_output.actions
    switch_beta = meta_output.switch_beta

    log_prob = meta_controller.log_prob(meta_output.action_dist, action)

    return GRPOOutput(state, action, log_prob, switch_beta)

@save_load()
class MetaController(Module):
    def __init__(
        self,
        dim_model,
        *,
        dim_meta_controller = 256,
        dim_latent = 128,
        switch_per_latent_dim = True,
        decoder_expansion_factor = 2.,
        decoder_depth = 1,
        hypernetwork_low_rank = 16,
        assoc_scan_kwargs: dict = dict(),
        bidirectional_temporal_encoder_kwargs: dict = dict(
            attn_dim_head = 32,
            heads = 8
        ),
        action_proposer: Module | dict = dict(
            depth = 1,
            attn_dim_head = 32,
            heads = 8
        )
    ):
        super().__init__()
        self.dim_model = dim_model
        dim_meta = default(dim_meta_controller, dim_model)

        # the linear that brings from model dimension 

        self.model_to_meta = Linear(dim_model, dim_meta)

        # there are two phases, the first (discovery ssl phase) uses acausal with some ssm i don't really believe in - let's just use bidirectional attention as placeholder

        self.bidirectional_temporal_encoder = Encoder(dim = dim_meta, depth = 1, **bidirectional_temporal_encoder_kwargs)

        self.emitter = GRU(dim_meta * 2, dim_meta * 2)
        self.emitter_to_action_mean_log_var = Readout(dim_meta * 2, num_continuous = dim_latent)

        if isinstance(action_proposer, dict):
            # default to Decoder, wrapped for standard interface

            action_proposer = ActionProposerWrapper(
                Decoder(dim = dim_meta, **action_proposer),
                cache_key = 'cache',
                return_cache_key = 'return_hiddens'
            )

        self.action_proposer = action_proposer
        self.action_proposer_mean_log_var = Readout(dim_meta, num_continuous = dim_latent)

        # switching unit

        self.switch_per_latent_dim = switch_per_latent_dim

        self.dim_latent = dim_latent
        self.switching_unit = GRU(dim_meta + dim_latent, dim_meta)
        self.to_switching_unit_beta = nn.Linear(dim_meta, dim_latent if switch_per_latent_dim else 1, bias = False)

        self.switch_gating = AssocScan(**assoc_scan_kwargs)

        # decoder

        assert hypernetwork_low_rank < dim_latent

        dim_decoder_hidden = int(dim_latent * decoder_expansion_factor)

        self.decoder = Feedforwards(
            dim_in = dim_latent,
            dim = dim_decoder_hidden,
            depth = decoder_depth,
            dim_out = 2 * hypernetwork_low_rank * dim_model
        )

        self.to_hyper_network_weights = Rearrange('... (two d r) -> two ... d r', two = 2, r = hypernetwork_low_rank)

        self.register_buffer('zero', tensor(0.), persistent = False)

    @property
    def replay_buffer_field_dict(self):
        return dict(
            states = ('float', self.dim_model),
            log_probs = ('float', self.dim_latent),
            switch_betas = ('float', self.dim_latent if self.switch_per_latent_dim else 1),
            latent_actions = ('float', self.dim_latent)
        )

    def discovery_parameters(self):
        return [
            *self.model_to_meta.parameters(),
            *self.bidirectional_temporal_encoder.parameters(),
            *self.emitter.parameters(),
            *self.emitter_to_action_mean_log_var.parameters(),
            *self.switching_unit.parameters(),
            *self.to_switching_unit_beta.parameters(),
            *self.decoder.parameters(),
            *self.switch_gating.parameters()
        ]

    def internal_rl_parameters(self):
        return [
            *self.action_proposer.parameters(),
            *self.action_proposer_mean_log_var.parameters()
        ]

    def get_action_dist_for_internal_rl(
        self,
        residual_stream
    ):
        meta_embed = self.model_to_meta(residual_stream)

        proposed_action_hidden, _ = self.action_proposer(meta_embed)

        return self.action_proposer_mean_log_var(proposed_action_hidden)

    def log_prob(
        self,
        action_dist,
        sampled_latent_action
    ):
        return self.action_proposer_mean_log_var.log_prob(action_dist, sampled_latent_action)

    def forward(
        self,
        residual_stream,
        cache: MetaControllerOutput | None = None,
        discovery_phase = False,
        hard_switch = None,
        temperature = 1.,
        episode_lens: Tensor | None = None
    ):
        device = residual_stream.device

        # destruct prev cache

        prev_action_proposer_hidden, prev_switching_unit_gru_hidden, prev_switch_gated_hiddens, prev_sampled_latent_action = cache.prev_hiddens if exists(cache) else ((None,) * 4)

        # getting proposed action for the two phases

        next_action_proposer_hidden = None

        meta_embed = self.model_to_meta(residual_stream)

        hard_switch = default(hard_switch, not discovery_phase) # think during internal RL phase, it needs to be a hard switch, then only the actions emitted during the switch is reinforced

        if discovery_phase:
            logger.warning('meta controller cache being passed back in for discovery phase, which does not make sense given bidirectional encoder')

            mask = maybe(lens_to_mask)(episode_lens, meta_embed.shape[1])

            encoded_temporal = self.bidirectional_temporal_encoder(meta_embed, mask = mask)

            proposed_action_hidden, _ = self.emitter(cat((encoded_temporal, meta_embed), dim = -1))
            readout = self.emitter_to_action_mean_log_var

        else: # else internal rl phase

            proposed_action_hidden, next_action_proposer_hidden = self.action_proposer(
                meta_embed,
                cache = prev_action_proposer_hidden
            )

            readout = self.action_proposer_mean_log_var

        # sample from the gaussian as the action from the meta controller

        action_dist = readout(proposed_action_hidden)

        sampled_latent_action = readout.sample(action_dist, temperature = temperature)

        # switching unit timer

        batch, seq_len, dim = sampled_latent_action.shape

        # initialize prev sampled latent action to be zeros if not available (for first timestep and for discovery phase)

        if not exists(prev_sampled_latent_action):
            prev_sampled_latent_action = torch.zeros(batch, 1, self.dim_latent, device = device)

        if discovery_phase:
            z_prev = cat((prev_sampled_latent_action, sampled_latent_action[:, :-1]), dim = 1)

        else:
            # else during inference, use the previous sampled latent action

            assert seq_len == 1, 'inference RL phase must be done one token at a time - if replaying for policy optimization, please use `get_action_dist_for_internal_rl`'
            z_prev = prev_sampled_latent_action

        # switch input is previous latent action and the embedding

        switch_input = torch.cat((meta_embed, z_prev), dim=-1)

        switching_unit_gru_out, next_switching_unit_gru_hidden = self.switching_unit(
            switch_input, 
            prev_switching_unit_gru_hidden
        )

        switch_beta = self.to_switching_unit_beta(switching_unit_gru_out).sigmoid()

        # need to encourage normal distribution

        kl_loss = switch_loss = self.zero

        if discovery_phase:
            mean, log_var = action_dist.unbind(dim = -1)

            kl_loss = (0.5 * (
                log_var.exp()
                + mean.square()
                - log_var
                - 1.
            ))

            kl_loss = kl_loss * switch_beta
            kl_loss = kl_loss.sum(dim = -1).mean()

            # encourage less switching

            switch_loss = switch_beta.mean()

        # maybe hard switch, then use associative scan

        if hard_switch:
            hard_switch_beta = (switch_beta > 0.5).float()
            switch_beta = straight_through(switch_beta, hard_switch_beta)

        forget_gate = 1. - switch_beta
        input_gate = switch_beta

        gated_action = self.switch_gating(forget_gate, sampled_latent_action * input_gate, prev = prev_switch_gated_hiddens)

        next_switch_gated_action = gated_action[:, -1]

        # decoder

        decoder_out = self.decoder(gated_action)

        w1, w2 = self.to_hyper_network_weights(decoder_out)
        hypernetwork_weight = einsum(w1, w2, '... i r, ... j r -> ... i j')

        # generating the residual stream controlling signal

        control_signal = einsum(residual_stream, hypernetwork_weight, '... d1, ... d1 d2 -> ... d1')

        # returning

        next_hiddens = (
            next_action_proposer_hidden,
            next_switching_unit_gru_hidden,
            next_switch_gated_action,
            sampled_latent_action[:, -1:]
        )

        # squeeze out the last dimension of switch_beta if single gate for all latent dimensions

        if not self.switch_per_latent_dim:
            switch_beta = rearrange(switch_beta, '... 1 -> ...')

        return control_signal, MetaControllerOutput(next_hiddens, residual_stream, action_dist, sampled_latent_action, switch_beta, kl_loss, switch_loss)

MetaController.policy_loss = policy_loss

# main transformer, which is subsumed into the environment after behavioral cloning

Hiddens = namedtuple('Hiddens', (
    'lower_body',
    'meta_controller',
    'upper_body'
))

TransformerOutput = namedtuple('TransformerOutput', (
    'residual_stream_latent',
    'prev_hiddens',
    'cache_steps'
))

@save_load()
class Transformer(Module):
    def __init__(
        self,
        dim,
        *,
        state_embed_readout: dict,
        action_embed_readout: dict,
        lower_body: Decoder | dict,
        upper_body: Decoder | dict,
        meta_controller: MetaController | None = None,
        dim_condition = None
    ):
        super().__init__()

        if exists(dim_condition):
            transformer_kwargs = dict(
                dim_condition = dim_condition,
                use_adaptive_rmsnorm = True,
                polar_pos_emb = True
            )
        else:
            transformer_kwargs = dict(
                use_rmsnorm = True,
                polar_pos_emb = True
            )

        if isinstance(lower_body, dict):
            lower_body = Decoder(dim = dim, pre_norm_has_final_norm = False, **transformer_kwargs, **lower_body)

        if isinstance(upper_body, dict):
            upper_body = Decoder(dim = dim, **transformer_kwargs, **upper_body)

        self.state_embed, self.state_readout = EmbedAndReadout(dim, **state_embed_readout)
        self.action_embed, self.action_readout = EmbedAndReadout(dim, **action_embed_readout)

        self.lower_body = lower_body
        self.upper_body = upper_body

        # polar positional embedding

        lower_attn_dim_head = lower_body.attn_dim_head
        lower_heads = lower_body.attn_heads

        upper_attn_dim_head = upper_body.attn_dim_head
        upper_heads = upper_body.attn_heads

        assert lower_attn_dim_head == upper_attn_dim_head
        assert lower_heads == upper_heads

        self.polar_pos_emb = PolarEmbedding(lower_attn_dim_head, heads = lower_heads)

        # meta controller

        self.meta_controller = meta_controller 

        self.register_buffer('zero', tensor(0.), persistent = False)

        # ensure devices match
        
        if exists(self.meta_controller): self._ensure_consistent_device(self.meta_controller)

    def _ensure_consistent_device(self, network):
        self.model_device = module_device(self)
        if module_device(network) != self.model_device:
            network.to(self.model_device)

    def evolve(
        self,
        num_generations,
        environment,
        **kwargs
    ):
        assert exists(self.meta_controller), '`meta_controller` must be passed in or defined on init for evolutionary strategies to be straightforwardly applied'

        evo_strat = EvoStrategy(
            self,
            num_generations = num_generations,
            environment = environment,
            params_to_optimize = self.meta_controller.internal_rl_parameters(),
            **kwargs
        )

        evo_strat()

    def forward(
        self,
        state,
        actions: Tensor | None = None,
        meta_controller: Module | None = None,
        cache: TransformerOutput | None = None,
        discovery_phase = False,
        force_behavior_cloning = False,
        meta_controller_temperature = 1.,
        return_raw_action_dist = False,
        return_latents = False,
        return_cache = False,
        episode_lens: Tensor | None = None,
        return_meta_controller_output = False,
        condition = None
    ):
        device = state.device

        # meta controller is either given or already given at init

        if exists(meta_controller): self._ensure_consistent_device(meta_controller)

        meta_controller = default(meta_controller, self.meta_controller)

        if force_behavior_cloning:
            assert not discovery_phase, 'discovery phase cannot be set to True if force behavioral cloning is set to True'
            meta_controller = None

        has_meta_controller = exists(meta_controller)

        assert not (discovery_phase and not has_meta_controller), 'meta controller must be made available during discovery phase'

        behavioral_cloning = force_behavior_cloning or (not has_meta_controller and not return_raw_action_dist)

        # by default, if meta controller is passed in, transformer is no grad

        lower_transformer_context = nullcontext if not has_meta_controller else torch.no_grad
        meta_controller_context = nullcontext if has_meta_controller else torch.no_grad
        upper_transformer_context = nullcontext if (not has_meta_controller or discovery_phase) else torch.no_grad

        # handle cache

        lower_transformer_hiddens, meta_hiddens, upper_transformer_hiddens = cache.prev_hiddens if exists(cache) else ((None,) * 3)

        cache_steps = cache.cache_steps if exists(cache) else 0

        # handle maybe behavioral cloning

        if behavioral_cloning or discovery_phase: # during behavior cloning and discovery phase, the network is predicting / reconstructing the next token

            assert exists(actions), f'`actions` cannot be empty when doing discovery or behavioral cloning'

            state, target_state = state[:, :-1], state[:, 1:]
            actions, target_actions = actions[:, :-1], actions[:, 1:]

            if exists(episode_lens):
                episode_lens = (episode_lens - 1).clamp(min = 0)

        # positional embedding

        seq_len = state.shape[1]

        pos = torch.arange(seq_len, device = device)

        polar_pos_emb = self.polar_pos_emb(pos)

        # transformer lower body

        with lower_transformer_context():

            state_embed = self.state_embed(state)

            # handle no past action for first timestep

            if exists(actions):
                action_embed = self.action_embed(actions)
            else:
                action_embed = state_embed[:, 0:0] # empty action embed

            if action_embed.shape[-2] == (state_embed.shape[-2] - 1):
                action_embed = pad_at_dim(action_embed, (1, 0), dim = 1)

            embed = state_embed + action_embed

            residual_stream, next_lower_hiddens = self.lower_body(
                embed,
                condition = condition,
                cache = lower_transformer_hiddens,
                polar_pos_emb = polar_pos_emb,
                seq_pos_offset = cache_steps,
                return_hiddens = True
            )

        # meta controller acts on residual stream here

        with meta_controller_context():

            if exists(meta_controller) and not behavioral_cloning:
                control_signal, next_meta_hiddens = meta_controller(residual_stream, cache = meta_hiddens, discovery_phase = discovery_phase, temperature = meta_controller_temperature, episode_lens = episode_lens)
            else:
                control_signal, next_meta_hiddens = self.zero, None

            modified_residual_stream = residual_stream + control_signal

        # modified residual stream sent back to transformer upper body

        with upper_transformer_context():

            attended, next_upper_hiddens = self.upper_body(
                modified_residual_stream,
                condition = condition,
                cache = upper_transformer_hiddens,
                polar_pos_emb = polar_pos_emb,
                return_hiddens = True
            )

            # head readout

            dist_params = self.action_readout(attended)

        # maybe return behavior cloning loss

        if behavioral_cloning:

            loss_mask = maybe(lens_to_mask)(episode_lens, state.shape[1])

            state_dist_params = self.state_readout(attended)
            state_clone_loss = self.state_readout.calculate_loss(state_dist_params, target_state, mask = loss_mask)

            action_clone_loss = self.action_readout.calculate_loss(dist_params, target_actions, mask = loss_mask)

            losses = BehavioralCloningLosses(state_clone_loss, action_clone_loss)

            if not return_meta_controller_output:
                return losses

            return losses, next_meta_hiddens

        elif discovery_phase:

            action_recon_loss = self.action_readout.calculate_loss(dist_params, target_actions)

            losses = DiscoveryLosses(action_recon_loss, next_meta_hiddens.kl_loss, next_meta_hiddens.switch_loss)

            if not return_meta_controller_output:
                return losses

            return losses, next_meta_hiddens

        # returning

        return_one = not (return_latents or return_cache)

        if return_cache:
            next_cache_steps = cache_steps + seq_len
            return_cache_value = TransformerOutput(residual_stream, Hiddens(next_lower_hiddens, next_meta_hiddens, next_upper_hiddens), next_cache_steps)

        if return_one:
            return dist_params

        return dist_params, return_cache_value



# /// script
# dependencies = [
#   "gymnasium",
#   "minigrid",
#   "tqdm",
#   "fire",
#   "memmap-replay-buffer>=0.0.23",
#   "loguru",
#   "sentence-transformers"
# ]
# ///

# taken with modifications from https://github.com/ddidacus/bot-minigrid-babyai/blob/main/tests/get_trajectories.py

import fire
import random
import multiprocessing
from loguru import logger
from pathlib import Path

import warnings
warnings.filterwarnings("ignore", category = UserWarning)

import numpy as np

from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor, wait, FIRST_COMPLETED

import torch
import minigrid
import gymnasium as gym
from babyai_env import get_missions_embeddings, get_sbert
from minigrid.utils.baby_ai_bot import BabyAIBot
from minigrid.wrappers import FullyObsWrapper, SymbolicObsWrapper

from gymnasium import spaces
from gymnasium.core import ObservationWrapper
from minigrid.core.constants import OBJECT_TO_IDX

from memmap_replay_buffer import ReplayBuffer

# Difficulty thresholds based on mission length
EASY_MAX_LENGTH = 30      # easy: 0 to 30
MEDIUM_MAX_LENGTH = 75    # medium: 30 to 75, hard: > 75

# helpers

def exists(val):
    return val is not None

def sample(prob):
    return random.random() < prob

def get_mission_for_seed(env_id, seed):
    """
    Get the mission string for a given seed.
    Returns (mission_length, mission_string).
    """
    env = gym.make(env_id, render_mode="rgb_array")
    env.reset(seed=seed)
    mission = env.unwrapped.mission
    env.close()
    return len(mission), mission

def get_mission_length(env_id, seed):
    """
    Get the mission length for a given seed.
    Returns the length of the mission string.
    """
    length, _ = get_mission_for_seed(env_id, seed)
    return length

def categorize_seeds_by_difficulty(env_id, num_seeds_per_level, level_difficulty=None):
    """
    Scan seeds and categorize them by difficulty based on mission length.
    Also collects the mission string for each seed.
    
    Args:
        env_id: Environment ID
        num_seeds_per_level: Number of seeds needed per difficulty level
        level_difficulty: List of levels to collect seeds for.
                       Supported: 'easy', 'medium', 'hard'
                       If None, collects for ['easy', 'hard'].
    
    Returns:
        tuple of:
            - dict with keys for each requested level, each containing a list of seeds
            - dict mapping seed -> mission string
    """
    
    seeds = {level: [] for level in level_difficulty}
    seed_to_mission = {}
    
    total_needed = sum(num_seeds_per_level for _ in level_difficulty)
    print(f"Scanning seeds to categorize by difficulty (need {num_seeds_per_level} per level for {level_difficulty})...")
    
    with tqdm(total=total_needed, desc="Categorizing seeds") as pbar:
        seed = 1
        all_done = False
        while not all_done:
            # Check if we have enough seeds for all requested levels
            all_done = all(len(seeds[level]) >= num_seeds_per_level for level in level_difficulty)
            
            try:
                mission_length, mission = get_mission_for_seed(env_id, seed)
                
                # easy: mission length <= 30
                if 'easy' in level_difficulty and mission_length <= EASY_MAX_LENGTH and len(seeds['easy']) < num_seeds_per_level:
                    seeds['easy'].append(seed)
                    seed_to_mission[seed] = mission
                    pbar.update(1)
                # medium: mission length <= 75 (combines easy and medium)
                elif 'medium' in level_difficulty and mission_length > EASY_MAX_LENGTH and mission_length <= MEDIUM_MAX_LENGTH and len(seeds['medium']) < num_seeds_per_level:
                    seeds['medium'].append(seed)
                    seed_to_mission[seed] = mission
                    pbar.update(1)
                # hard: mission length > 75
                elif 'hard' in level_difficulty and mission_length > MEDIUM_MAX_LENGTH and len(seeds['hard']) < num_seeds_per_level:
                    seeds['hard'].append(seed)
                    seed_to_mission[seed] = mission
                    pbar.update(1)
            except Exception as e:
                logger.warning(f"Error getting mission for seed {seed}: {e}")
            
            seed += 1
    
    return seeds, seed_to_mission

# wrapper, necessarily modified to allow for both rgb obs (policy) and symbolic obs (bot)

class RGBImgPartialObsWrapper(ObservationWrapper):
    """
    Wrapper to use partially observable RGB image as observation.
    This can be used to have the agent to solve the gridworld in pixel space.
    """
    def __init__(self, env, tile_size=1):
        super().__init__(env)

        # Rendering attributes for observations
        self.tile_size = tile_size

        symbolic_image_space = self.observation_space["image"]

        obs_shape = env.observation_space.spaces["image"].shape
        new_image_space = spaces.Box(
            low=0,
            high=255,
            shape=(obs_shape[0] * tile_size, obs_shape[1] * tile_size, 3),
            dtype="uint8",
        )

        self.observation_space = spaces.Dict(
            {**self.observation_space.spaces, "image": symbolic_image_space, "rgb_image": new_image_space}
        )

    def observation(self, obs):
        rgb_img_partial = self.unwrapped.get_frame(
            tile_size=self.tile_size, agent_pov=True
        )

        return {**obs, "rgb_image": rgb_img_partial}

# agent

class BabyAIBotEpsilonGreedy:
    def __init__(self, env, random_action_prob = 0.):
        self.expert = BabyAIBot(env)
        self.random_action_prob = random_action_prob
        self.num_actions = env.action_space.n
        self.last_action = None

    def __call__(self, state):
        if sample(self.random_action_prob):
            action = torch.randint(0, self.num_actions, ()).item()
        else:
            action = self.expert.replan(self.last_action)

        self.last_action = action
        return action

# functions

def collect_single_episode(env_id, seed, num_steps, random_action_prob, state_shape):
    """
    Collect a single episode of demonstrations.
    Returns tuple of (episode_state, episode_action, success, episode_length, seed)
    Note: mission embeddings are pre-computed in main process and looked up by seed.
    """
    if env_id not in gym.envs.registry:
        minigrid.register_minigrid_envs()

    env = gym.make(env_id, render_mode="rgb_array", highlight=False)
    env = FullyObsWrapper(env.unwrapped)
    env = SymbolicObsWrapper(env.unwrapped)
    env = RGBImgPartialObsWrapper(env.unwrapped)

    try:
        state_obs, _ = env.reset(seed=seed)
        episode_state = np.zeros((num_steps, *state_shape), dtype=np.float32)
        episode_action = np.zeros(num_steps, dtype=np.float32)

        expert = BabyAIBotEpsilonGreedy(env.unwrapped, random_action_prob=random_action_prob)

        for _step in range(num_steps):
            try:
                action = expert(state_obs)
            except Exception:
                env.close()
                return None, None, False, 0, seed

            episode_state[_step] = state_obs["rgb_image"]
            episode_action[_step] = action

            state_obs, reward, terminated, truncated, info = env.step(action)

            if terminated:
                env.close()
                return episode_state, episode_action, True, _step + 1, seed

        env.close()
        return episode_state, episode_action, False, num_steps, seed

    except Exception:
        env.close()
        return None, None, False, 0, seed

def collect_demonstrations(
    env_id = "BabyAI-MiniBossLevel-v0",
    num_seeds = 100,
    num_episodes_per_seed = 100,
    num_steps = 500,
    random_action_prob = 0.05,
    num_workers = None,
    difficulty = "easy",
    output_dir = "babyai-minibosslevel-trajectories",
    mission_embed_dim = 384
):
    """
    The BabyAI Bot should be able to solve all BabyAI environments,
    allowing us therefore to generate demonstrations.
    Parallelized version using ProcessPoolExecutor.
    """

    # Register minigrid envs if not already registered
    if env_id not in gym.envs.registry:
        minigrid.register_minigrid_envs()

    # Determine state shape from environment
    temp_env = gym.make(env_id)
    temp_env = FullyObsWrapper(temp_env.unwrapped)
    temp_env = SymbolicObsWrapper(temp_env.unwrapped)
    temp_env = RGBImgPartialObsWrapper(temp_env.unwrapped)
    state_shape = temp_env.observation_space['rgb_image'].shape
    temp_env.close()

    logger.info(f"Detected state shape: {state_shape} for env {env_id}")

    if not exists(num_workers):
        num_workers = multiprocessing.cpu_count()

    total_episodes = num_seeds * num_episodes_per_seed

    # filter seeds

    assert difficulty in ['easy', 'medium', 'hard']
    seeds, seed_to_mission = categorize_seeds_by_difficulty(env_id, num_seeds_per_level=num_seeds, level_difficulty=[difficulty])

    # pre-compute embeddings on master process (harder to handle with subprocesses)

    logger.info("Pre-computing mission embeddings with GPU...")
    sbert = get_sbert(local_files_only = False, device = "cuda" if torch.cuda.is_available() else "cpu")
    selected_seeds = seeds[difficulty]
    missions = [seed_to_mission[s] for s in selected_seeds]
    embeddings = get_missions_embeddings(missions, sbert).cpu().numpy()
    seed_to_embedding = {s: embeddings[i] for i, s in enumerate(selected_seeds)}

    successful = 0
    progressbar = tqdm(total=total_episodes)

    output_folder = Path(output_dir)

    fields = {
        'state': ('float', state_shape),
        'action': ('float', ())
    }

    meta_fields = {
        'mission_embedding': ('float', (mission_embed_dim,))
    }

    buffer = ReplayBuffer(
        folder = output_folder,
        max_episodes = total_episodes,
        max_timesteps = num_steps,
        fields = fields,
        meta_fields = meta_fields,
        overwrite = True
    )

    max_pending = num_workers
    all_seeds = seeds[difficulty] * num_episodes_per_seed

    with ProcessPoolExecutor(max_workers=num_workers) as executor:
        seed_iter = iter(all_seeds)
        futures = {}

        # submit

        for _ in range(min(max_pending, len(all_seeds))):
            seed = next(seed_iter, None)
            if exists(seed):
                future = executor.submit(collect_single_episode, env_id, seed, num_steps, random_action_prob, state_shape)
                futures[future] = seed

        # collect

        while futures:
            done, _ = wait(futures, return_when=FIRST_COMPLETED)

            for future in done:
                seed = futures.pop(future)
                episode_state, episode_action, success, episode_length, returned_seed = future.result()

                if success and exists(episode_state):

                    # get cached mission embedding (forced to exist)

                    mission_embedding = seed_to_embedding[returned_seed]
                    buffer.store_episode(
                        state = episode_state[:episode_length],
                        action = episode_action[:episode_length],
                        mission_embedding = mission_embedding
                    )
                    successful += 1

                progressbar.update(1)
                progressbar.set_description(f"success rate = {successful}/{progressbar.n:.2f}")

                seed = next(seed_iter, None)
                if exists(seed):
                    new_future = executor.submit(collect_single_episode, env_id, seed, num_steps, random_action_prob, state_shape)
                    futures[new_future] = seed

    buffer.flush()
    progressbar.close()

    seeds_array = np.array(seeds[difficulty])
    seeds_path = output_folder / "seeds.npy"
    np.save(seeds_path, seeds_array)

    logger.info(f"Saved {successful} trajectories to {output_dir}")
    logger.info(f"Saved {len(seeds_array)} seeds to {seeds_path}")

if __name__ == "__main__":
    fire.Fire(collect_demonstrations)



# /// script
# dependencies = [
#   "fire",
#   "gymnasium",
#   "gymnasium[other]",
#   "metacontroller-pytorch>=0.0.57",
#   "torch-einops-utils>=0.0.27",
#   "minigrid",
#   "tqdm",
#   "wandb",
#   "sentence-transformers"
# ]
# ///
# To train with multiple GPUs:
# 1. accelerate config
# 2. accelerate launch train_babyai_onpolicy.py [args]

from fire import Fire
from pathlib import Path
from tqdm import tqdm

import numpy as np
import torch
from torch import cat, tensor, stack, Tensor
from torch.optim import Adam

from einops import rearrange

from torch_einops_utils import pad_sequence, pad_sequence_and_cat, lens_to_mask

from accelerate import Accelerator

from babyai_env import (
    create_env,
    get_missions_embeddings,
    transform_to_symbolic
)
import gymnasium as gym
from gymnasium.vector import AsyncVectorEnv
from metacontroller.metacontroller import Transformer, MetaController, z_score, extract_grpo_data
from metacontroller.transformer_with_resnet import TransformerWithResnet
from functools import partial

# Patch x_transformers so checkpoint config can unpickle (Identity was removed/moved in some versions)
try:
    import x_transformers.x_transformers as _xt_module
    if not hasattr(_xt_module, "Identity"):
        _xt_module.Identity = torch.nn.Identity
except Exception:
    pass

# research entry point

def reward_shaping_fn(
    cumulative_rewards: Tensor, # float(num_episodes,)
    all_rewards: Tensor,        # float(num_episodes, max_timesteps)
    episode_lens: Tensor,       # int(num_episodes,)
    reject_threshold_cumulative_reward_variance: float = 0.
) -> Tensor | None:
    """
    researchers can modify this function to engineer rewards
    or return None to reject the entire batch
    """

    if cumulative_rewards.var() < reject_threshold_cumulative_reward_variance:
        return None

    return cumulative_rewards

def should_reject_group_based_on_switch_betas(
    switch_betas: Tensor,
    episode_lens: Tensor
):
    return switch_betas.sum().item() == 0.

# helpers

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

def divisible_by(num, den):
    return (num % den) == 0


# main

def main(
    seed: int | None = None,
    npy_seedfile = None,
    env_name = 'BabyAI-BossLevel-v0',
    num_episodes = int(10e6),
    max_timesteps = 500,
    render_every_eps = 1_000,
    video_folder = './recordings',
    transformer_weights_path: str | None = None,
    meta_controller_weights_path: str | None = None,
    output_meta_controller_path = 'metacontroller_rl_trained.pt',
    use_resnet = False,
    lr = 3e-5,
    save_steps = 100,
    batch_size = 16,
    max_grad_norm = 1.0,
    use_wandb = False,
    wandb_project = 'metacontroller-babyai-rl',
    reject_threshold_cumulative_reward_variance = 0.1,
    condition_on_mission_embed = False,
    env_shared_memory = True,
    env_context = 'fork'
):

    def store_checkpoint(step: int):
        if accelerator.is_main_process:
            meta_controller_checkpoint_path_with_step = output_meta_controller_path.replace('.pt', f'_step_{step}.pt')
            unwrapped_meta_controller = accelerator.unwrap_model(meta_controller)
            unwrapped_meta_controller.save(meta_controller_checkpoint_path_with_step)
            accelerator.print(f"MetaController to {meta_controller_checkpoint_path_with_step}")

    # seed selection priority: 1) fixed seed arg, 2) numpy seedfile, 3) random

    group_seeds = None
    if not exists(seed) and exists(npy_seedfile):
        cap_seeds = 10
        group_seeds = np.load(npy_seedfile, allow_pickle=True).astype(int).tolist()
        group_seeds = group_seeds[:cap_seeds]
        
    seed_index = 0

    def get_next_seed():
        nonlocal seed_index
        if exists(seed):
            return seed
        if group_seeds is not None:
            s = group_seeds[seed_index % len(group_seeds)]
            seed_index += 1
            return s
        return torch.randint(0, 1000000, (1,)).item()

    # batch_size = number of simultaneous trajectories (the group) = number of parallel envs
    assert batch_size >= 2, "batch_size must be at least 2 for relative comparison (GRPO)"

    # accelerator

    accelerator = Accelerator(log_with = 'wandb' if use_wandb else None)

    if use_wandb:
        accelerator.init_trackers(wandb_project)

    # environment

    env_make_fn = partial(
        create_env,
        env_name,
        render_mode = 'rgb_array',
        video_folder = video_folder,
        render_every_eps = render_every_eps,
        use_symbolic = False
    )

    env = AsyncVectorEnv([env_make_fn] * batch_size, shared_memory = env_shared_memory, context = env_context)


    # load models

    model = None
    if exists(transformer_weights_path):
        weights_path = Path(transformer_weights_path)
        assert weights_path.exists(), f"transformer weights not found at {weights_path}"
        
        transformer_klass = TransformerWithResnet if use_resnet else Transformer

        model = transformer_klass.init_and_load(str(weights_path), strict = False)
        
        model.eval()

    meta_controller = None
    if exists(meta_controller_weights_path):
        weights_path = Path(meta_controller_weights_path)
        assert weights_path.exists(), f"meta controller weights not found at {weights_path}"
        meta_controller = MetaController.init_and_load(str(weights_path), strict = False)
        meta_controller.eval()

    meta_controller = default(meta_controller, getattr(model, 'meta_controller', None))
    assert exists(meta_controller), "MetaController must be present for reinforcement learning"

    # optimizer

    optim = Adam(meta_controller.internal_rl_parameters(), lr = lr)

    # prepare

    model, meta_controller, optim = accelerator.prepare(model, meta_controller, optim)

    unwrapped_model = accelerator.unwrap_model(model) if exists(model) else None
    unwrapped_meta_controller = accelerator.unwrap_model(meta_controller)

    # rollouts

    num_batch_updates = num_episodes // batch_size

    pbar = tqdm(range(num_batch_updates), desc = 'training')
    
    print("starting training")

    for gradient_step in pbar:

        # one group of batch_size trajectories per gradient step (for GRPO relative comparison)
        group_seed = get_next_seed()
        env_seeds = [group_seed] * batch_size

        state, _ = env.reset(seed = env_seeds)

        if condition_on_mission_embed:
            missions = env.get_attr('mission')
            mission_embed = get_missions_embeddings(missions)
            mission_embed = mission_embed.to(accelerator.device)

        cache = None
        past_action_id = None

        iteration_states = []
        iteration_log_probs = []
        iteration_switch_betas = []
        iteration_latent_actions = []
        iteration_rewards = []

        dones = torch.zeros(batch_size, dtype = torch.bool)

        all_steps_dones = []

        # rollout: batch of batch_size parallel trajectories (one env per trajectory)
        for step in range(max_timesteps):
            # state['image']: (batch_size, H, W, C)
            image = state['image']
            image_tensor = torch.from_numpy(image).float().to(accelerator.device)

            if use_resnet:
                image_tensor = rearrange(image_tensor, 'b h w c -> b 1 h w c')
                image_tensor = unwrapped_model.visual_encode(image_tensor)
            else:
                image_tensor = rearrange(image_tensor, 'b h w c -> b 1 (h w c)')

            if torch.is_tensor(past_action_id):
                past_action_id = past_action_id.long()

            with torch.no_grad():
                logits, cache = unwrapped_model(
                    image_tensor,
                    past_action_id,
                    meta_controller = unwrapped_meta_controller,
                    return_cache = True,
                    return_raw_action_dist = True,
                    cache = cache,
                    condition = mission_embed if condition_on_mission_embed else None
                )

            action = unwrapped_model.action_readout.sample(logits)
            past_action_id = action

            grpo_data = extract_grpo_data(unwrapped_meta_controller, cache)

            iteration_states.append(grpo_data.state)
            iteration_log_probs.append(grpo_data.log_prob)
            iteration_switch_betas.append(grpo_data.switch_beta)
            iteration_latent_actions.append(grpo_data.action)

            next_state, reward, terminated, truncated, _ = env.step(action.cpu().numpy())
            next_state['image'] = transform_to_symbolic(next_state['image'])

            reward_tensor = torch.from_numpy(reward).float().to(accelerator.device)
            iteration_rewards.append(rearrange(reward_tensor, 'b -> b 1'))

            all_steps_dones.append(dones.clone())
            dones |= torch.from_numpy(terminated | truncated)

            if dones.all():
                break

            state = next_state

        episode_lens = (~torch.stack(all_steps_dones)).sum(dim = 0).to(accelerator.device)

        cur_states = cat(iteration_states, dim = 1)
        cur_log_probs = cat(iteration_log_probs, dim = 1)
        cur_switch_betas = cat(iteration_switch_betas, dim = 1)
        cur_latent_actions = cat(iteration_latent_actions, dim = 1)
        cur_rewards = cat(iteration_rewards, dim = 1)

        # pack group data (single rollout → one group of batch_size trajectories)

        group_states = cur_states
        group_log_probs = cur_log_probs
        group_switch_betas = cur_switch_betas
        group_latent_actions = cur_latent_actions
        
        group_step_rewards = cur_rewards

        # mask rewards after done

        mask = lens_to_mask(episode_lens, group_step_rewards.shape[-1])
        group_step_rewards = group_step_rewards * mask

        # compute advantages via z-score (GRPO style)

        cumulative_rewards = group_step_rewards.sum(dim = -1)

        # reward shaping hook

        shaped_rewards = reward_shaping_fn(
            cumulative_rewards,
            group_step_rewards,
            episode_lens,
            reject_threshold_cumulative_reward_variance = reject_threshold_cumulative_reward_variance
        )

        if not exists(shaped_rewards):
            accelerator.print(f'group rejected - variance of {cumulative_rewards.var().item():.4f} is lower than threshold of {reject_threshold_cumulative_reward_variance}')
            continue

        group_advantages = z_score(shaped_rewards).float()

        if torch.any(torch.isnan(group_advantages)):
            accelerator.print(f'group rejected - advantages contained NaNs')
            continue

        # whether to reject group based on switch betas (as it determines the mask for learning)

        if should_reject_group_based_on_switch_betas(group_switch_betas, episode_lens):
            accelerator.print(f'group rejected - switch betas for the entire group does not meet criteria for learning')
            continue

        # learn on this group directly (on-policy GRPO)

        meta_controller.train()

        loss = meta_controller.policy_loss(
            group_states,
            group_log_probs,
            group_latent_actions,
            group_advantages,
            group_switch_betas == 1.,
            episode_lens = episode_lens
        )

        accelerator.backward(loss)

        grad_norm = accelerator.clip_grad_norm_(meta_controller.parameters(), max_grad_norm)

        optim.step()
        optim.zero_grad()

        meta_controller.eval()

        pbar.set_postfix(
            loss = f'{loss.item():.4f}',
            grad_norm = f'{grad_norm.item():.4f}',
            reward = f'{cumulative_rewards.mean().item():.4f}',
            seeds = f'{group_seeds}'
        )

        accelerator.log({
            'loss': loss.item(),
            'grad_norm': grad_norm.item(),
            'reward': cumulative_rewards.mean().item(),
            'reward_std': cumulative_rewards.std().item(),
        })

        accelerator.print(f'loss: {loss.item():.4f}, grad_norm: {grad_norm.item():.4f}, reward: {cumulative_rewards.mean().item():.4f}, seeds: {group_seeds}')

        if gradient_step % save_steps == 0:
            store_checkpoint(gradient_step)

    env.close()

    # save

    if exists(output_meta_controller_path):
        unwrapped_meta_controller.save(output_meta_controller_path)
        accelerator.print(f'MetaController weights saved to {output_meta_controller_path}')

if __name__ == '__main__':
    Fire(main)



# /// script
# dependencies = [
#   "fire",
#   "gymnasium",
#   "gymnasium[other]",
#   "memmap-replay-buffer>=0.0.12",
#   "metacontroller-pytorch>=0.0.57",
#   "minigrid",
#   "tqdm",
#   "wandb",
#   "sentence-transformers"
# ]
# ///
# To train with multiple GPUs:
# 1. accelerate config
# 2. accelerate launch train_babyai.py [args]

from fire import Fire
from pathlib import Path
from functools import partial
from shutil import rmtree
from tqdm import tqdm

import numpy as np
import torch
from torch import cat, tensor, stack, Tensor
from torch.optim import Adam

from einops import rearrange

from torch_einops_utils import maybe, pad_at_dim, lens_to_mask, masked_mean, align_dims_left, pad_sequence, pad_sequence_and_cat

from accelerate import Accelerator

import gymnasium as gym
from gymnasium.vector import AsyncVectorEnv

from babyai_env import (
    create_env,
    get_missions_embeddings,
    transform_to_symbolic
)

from memmap_replay_buffer import ReplayBuffer

from metacontroller.metacontroller import Transformer, MetaController, policy_loss, z_score, extract_grpo_data
from metacontroller.transformer_with_resnet import TransformerWithResnet

# research entry point

def reward_shaping_fn(
    cumulative_rewards: Tensor, # float(num_episodes,)
    all_rewards: Tensor,        # float(num_episodes, max_timesteps)
    episode_lens: Tensor,       # int(num_episodes,)
    reject_threshold_cumulative_reward_variance: float = 0.
) -> Tensor | None:
    """
    researchers can modify this function to engineer rewards
    or return None to reject the entire batch
    """

    if cumulative_rewards.var() < reject_threshold_cumulative_reward_variance:
        return None

    return cumulative_rewards

def should_reject_group_based_on_switch_betas(
    switch_betas: Tensor,
    episode_lens: Tensor
):
    return switch_betas.sum().item() == 0.

# helpers

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

def divisible_by(num, den):
    return (num % den) == 0


# main

def main(
    npy_seedfile = None,
    env_name = 'BabyAI-BossLevel-v0',
    gradient_accumulation_steps = None,
    num_episodes = int(10e6),
    max_timesteps = 500,
    buffer_size = 1_000,
    render_every_eps = 1_000,
    video_folder = './recordings',
    seed: int | None = None,
    transformer_weights_path: str | None = None,
    meta_controller_weights_path: str | None = None,
    output_meta_controller_path = 'metacontroller_rl_trained.pt',
    use_resnet = False,
    num_epochs = 3,
    lr = 1e-4,
    save_steps = 100,
    batch_size = 16,
    num_groups = 16,
    max_grad_norm = 1.0,
    use_wandb = False,
    wandb_project = 'metacontroller-babyai-rl',
    reject_threshold_cumulative_reward_variance = 0.,
    condition_on_mission_embed = False,
    num_envs = 1,
    env_shared_memory = False,
    env_context = 'spawn'
):

    def store_checkpoint(step:int):
        if accelerator.is_main_process:
            # Add step to checkpoint filenames
            meta_controller_checkpoint_path_with_step = output_meta_controller_path.replace('.pt', f'_step_{step}.pt')
            unwrapped_meta_controller = accelerator.unwrap_model(meta_controller)
            unwrapped_meta_controller.save(meta_controller_checkpoint_path_with_step)
            accelerator.print(f"MetaController to {meta_controller_checkpoint_path_with_step}")

    # seed selection priority: 1) fixed seed arg, 2) numpy seedfile, 3) random

    group_seeds = None
    if not exists(seed) and exists(npy_seedfile):
        group_seeds = np.load(npy_seedfile, allow_pickle=True).astype(int).tolist()
        
    seed_index = 0

    def get_next_seed():
        nonlocal seed_index
        if exists(seed):
            return seed
        if group_seeds is not None:
            s = group_seeds[seed_index % len(group_seeds)]
            seed_index += 1
            return s
        return torch.randint(0, 1000000, (1,)).item()

    # check num_envs

    assert divisible_by(num_envs, num_groups) or divisible_by(num_groups, num_envs), f"one of num_groups ({num_groups}) or num_envs ({num_envs}) must be divisible by the other"
    assert num_groups >= 2, "num_groups must be at least 2 for relative comparison (GRPO)"

    # accelerator

    accelerator = Accelerator(log_with = 'wandb' if use_wandb else None)

    if use_wandb:
        accelerator.init_trackers(wandb_project)

    # environment

    # environment

    env_make_fn = partial(
        create_env,
        env_name,
        render_mode = 'rgb_array',
        video_folder = video_folder,
        render_every_eps = render_every_eps,
        use_symbolic = False
    )

    env = AsyncVectorEnv([env_make_fn] * num_envs, shared_memory = env_shared_memory, context = env_context)



    # load models

    model = None
    if exists(transformer_weights_path):
        weights_path = Path(transformer_weights_path)
        assert weights_path.exists(), f"transformer weights not found at {weights_path}"
        
        transformer_klass = TransformerWithResnet if use_resnet else Transformer
        model = transformer_klass.init_and_load(str(weights_path), strict = False)
        model.eval()

    meta_controller = None
    if exists(meta_controller_weights_path):
        weights_path = Path(meta_controller_weights_path)
        assert weights_path.exists(), f"meta controller weights not found at {weights_path}"
        meta_controller = MetaController.init_and_load(str(weights_path), strict = False)
        meta_controller.eval()

    meta_controller = default(meta_controller, getattr(model, 'meta_controller', None))
    assert exists(meta_controller), "MetaController must be present for reinforcement learning"

    # optimizer

    optim = Adam(meta_controller.internal_rl_parameters(), lr = lr)

    # prepare

    model, meta_controller, optim = accelerator.prepare(model, meta_controller, optim)

    unwrapped_model = accelerator.unwrap_model(model) if exists(model) else None
    unwrapped_meta_controller = accelerator.unwrap_model(meta_controller)

    # replay buffer

    replay_buffer = ReplayBuffer(
        './replay-data',
        max_episodes = buffer_size,
        max_timesteps = max_timesteps + 1,
        fields = unwrapped_meta_controller.replay_buffer_field_dict,
        meta_fields = dict(advantages = 'float'),
        overwrite = True,
        circular = True
    )

    # rollouts

    num_batch_updates = num_episodes // num_groups

    num_rollout_iterations = max(1, num_groups // num_envs)
    num_groups_per_iteration = max(1, num_envs // num_groups)


    pbar = tqdm(range(num_batch_updates), desc = 'training')

    gradient_step = 0

    for _ in pbar:

        all_states = []
        all_log_probs = []
        all_switch_betas = []
        all_latent_actions = []
        all_step_rewards = []
        all_episode_lens = []

        # seeds for each group (for GRPO relative comparison)

        if num_envs <= num_groups:
            group_seeds = [get_next_seed()]
            env_seeds = group_seeds * num_envs
        else:
            group_seeds = [get_next_seed() for _ in range(num_groups_per_iteration)]
            env_seeds = [s for s in group_seeds for _ in range(num_groups)]

        for _ in range(num_rollout_iterations):

            state, _ = env.reset(seed = env_seeds)
            state['image'] = transform_to_symbolic(state['image'])

            if condition_on_mission_embed:
                missions = env.get_attr('mission')
                mission_embed = get_missions_embeddings(missions)
                mission_embed = mission_embed.to(accelerator.device)

            cache = None
            past_action_id = None

            iteration_states = []
            iteration_log_probs = []
            iteration_switch_betas = []
            iteration_latent_actions = []
            iteration_rewards = []

            dones = torch.zeros(num_envs, dtype = torch.bool)

            all_steps_dones = []

            # rollout parallel trajectories

            for step in range(max_timesteps):

                image = state['image']
                image_tensor = torch.from_numpy(image).float().to(accelerator.device)

                if use_resnet:
                    image_tensor = rearrange(image_tensor, 'b h w c -> b 1 h w c')
                    image_tensor = unwrapped_model.visual_encode(image_tensor)
                else:
                    image_tensor = rearrange(image_tensor, 'b h w c -> b 1 (h w c)')

                if torch.is_tensor(past_action_id):
                    past_action_id = past_action_id.long()

                with torch.no_grad():
                    logits, cache = unwrapped_model(
                        image_tensor,
                        past_action_id,
                        meta_controller = unwrapped_meta_controller,
                        return_cache = True,
                        return_raw_action_dist = True,
                        cache = cache,
                        condition = mission_embed if condition_on_mission_embed else None
                    )

                action = unwrapped_model.action_readout.sample(logits)
                past_action_id = action

                # GRPO collection

                grpo_data = extract_grpo_data(unwrapped_meta_controller, cache)

                iteration_states.append(grpo_data.state)
                iteration_log_probs.append(grpo_data.log_prob)
                iteration_switch_betas.append(grpo_data.switch_beta)
                iteration_latent_actions.append(grpo_data.action)

                next_state, reward, terminated, truncated, _ = env.step(action.cpu().numpy())
                next_state['image'] = transform_to_symbolic(next_state['image'])

                reward_tensor = torch.from_numpy(reward).float().to(accelerator.device)
                iteration_rewards.append(rearrange(reward_tensor, 'b -> b 1'))

                all_steps_dones.append(dones.clone())
                dones |= torch.from_numpy(terminated | truncated)

                if dones.all():
                    break

                state = next_state

            # post-process for each environment

            episode_lens = (~torch.stack(all_steps_dones)).sum(dim = 0)

            cur_states = cat(iteration_states, dim = 1)
            cur_log_probs = cat(iteration_log_probs, dim = 1)
            cur_switch_betas = cat(iteration_switch_betas, dim = 1)
            cur_latent_actions = cat(iteration_latent_actions, dim = 1)
            cur_rewards = cat(iteration_rewards, dim = 1)

            # store environment data in batches per iteration

            all_states.append(cur_states)
            all_log_probs.append(cur_log_probs)
            all_switch_betas.append(cur_switch_betas)
            all_latent_actions.append(cur_latent_actions)
            all_step_rewards.append(cur_rewards)
            all_episode_lens.append(episode_lens)

        # align all sequences across rollout iterations
 
        group_states = pad_sequence_and_cat(all_states, dim = 1, dim_cat = 0)
        group_log_probs = pad_sequence_and_cat(all_log_probs, dim = 1, dim_cat = 0)
        group_switch_betas = pad_sequence_and_cat(all_switch_betas, dim = 1, dim_cat = 0)
        group_latent_actions = pad_sequence_and_cat(all_latent_actions, dim = 1, dim_cat = 0)
        
        group_step_rewards = pad_sequence_and_cat(all_step_rewards, dim = 0, dim_cat = 1)

        episode_lens = cat(all_episode_lens, dim = 0).to(accelerator.device)
        
        # mask rewards after done

        mask = lens_to_mask(episode_lens, group_step_rewards.shape[-1])
        group_step_rewards = group_step_rewards * mask

        cumulative_rewards = group_step_rewards.sum(dim = -1)

        # reward shaping hook

        shaped_rewards = reward_shaping_fn(
            cumulative_rewards,
            group_step_rewards,
            episode_lens,
            reject_threshold_cumulative_reward_variance = reject_threshold_cumulative_reward_variance
        )

        if not exists(shaped_rewards):
            accelerator.print(f'group rejected - variance of {cumulative_rewards.var().item():.4f} is lower than threshold of {reject_threshold_cumulative_reward_variance}')
            continue

        grouped_shaped_rewards = rearrange(shaped_rewards, '(g n) -> g n', n = num_groups)
        grouped_advantages = z_score(grouped_shaped_rewards, dim = 1)
        group_advantages = rearrange(grouped_advantages, 'g n -> (g n)').float()

        # whether to reject group based on switch betas (as it determines the mask for learning)

        if should_reject_group_based_on_switch_betas(group_switch_betas, episode_lens):
            accelerator.print(f'group rejected - switch betas for the entire group does not meet criteria for learning')
            continue

        num_trajectories_collected = max(num_groups, num_envs)

        for i in range(num_trajectories_collected):
            L = episode_lens[i]
            replay_buffer.store_episode(
                states = group_states[i, :L],
                log_probs = group_log_probs[i, :L],
                switch_betas = group_switch_betas[i, :L],
                latent_actions = group_latent_actions[i, :L],
                advantages = group_advantages[i]
            )

        # learn

        if len(replay_buffer) >= buffer_size:
            dl = replay_buffer.dataloader(batch_size = batch_size, shuffle = True)
            dl = accelerator.prepare(dl)

            meta_controller.train()

            for epoch in range(num_epochs):
                for batch in dl:
                    loss = meta_controller.policy_loss(
                        batch['states'],
                        batch['log_probs'],
                        batch['latent_actions'],
                        batch['advantages'],
                        batch['switch_betas'] == 1.,
                        episode_lens = batch['_lens']
                    )

                    # gradient acc

                    if gradient_accumulation_steps != None: loss /= gradient_accumulation_steps

                    accelerator.backward(loss)

                    if gradient_accumulation_steps == None or gradient_step % gradient_accumulation_steps == 0:

                        grad_norm = accelerator.clip_grad_norm_(meta_controller.parameters(), max_grad_norm)

                        optim.step()
                        optim.zero_grad()

                        pbar.set_postfix(
                            epoch = epoch,
                            loss = f'{loss.item():.4f}',
                            grad_norm = f'{grad_norm.item():.4f}',
                            reward = f'{cumulative_rewards.mean().item():.4f}'
                        )

                        accelerator.log({
                            'loss': loss.item(),
                            'grad_norm': grad_norm.item(),
                            'reward': cumulative_rewards.mean().item()
                        })

                    # checkpointing
                    
                    if gradient_step % save_steps == 0:
                        store_checkpoint(gradient_step)

            meta_controller.eval()

            accelerator.print(f'loss: {loss.item():.4f}, grad_norm: {grad_norm.item():.4f}, reward: {cumulative_rewards.mean().item():.4f}')

    env.close()

    # save

    if exists(output_meta_controller_path):
        unwrapped_meta_controller.save(output_meta_controller_path)
        accelerator.print(f'MetaController weights saved to {output_meta_controller_path}')

if __name__ == '__main__':
    Fire(main)



# /// script
# dependencies = [
#   "fire",
#   "gymnasium",
#   "gymnasium[other]",
#   "metacontroller-pytorch>=0.0.57",
#   "minigrid",
#   "tqdm",
#   "x-evolution",
#   "einops",
#   "sentence-transformers",
#   "accelerate"
# ]
# ///
# To train with multiple GPUs:
# 1. accelerate config
# 2. accelerate launch train_babyai_evo_strat.py [args]

from __future__ import annotations
import fire
from pathlib import Path
from shutil import rmtree
import numpy as np

import torch
from torch import nn, Tensor, tensor
from torch.nn import Module
from einops import rearrange

from babyai_env import create_env, get_mission_embedding
from metacontroller.metacontroller import Transformer, MetaController
from torch_einops_utils.device import module_device

from accelerate import Accelerator

# functions

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

# default fitness function

def default_fitness_fn(
    rewards: list[float],
    states: list[any],
    actions: list[any],
    next_states: list[any],
    infos: list[any]
) -> float:
    """
    researchers can modify this function to engineer their own rewards and fitness scores
    processing the entire episode at once for every noise vector of the population separately
    """
    return sum(rewards)

# babyai environment for ES

class BabyAIEnvironment(Module):
    def __init__(
        self,
        env_id = 'BabyAI-BossLevel-v0',
        video_folder = './recordings_babyai_es',
        render_every_eps = 100,
        max_steps = 500,
        use_resnet = False,
        condition_on_mission_embed = False,
        fitness_fn = default_fitness_fn,
        accelerator: Accelerator | None = None
    ):
        super().__init__()
        self.accelerator = accelerator

        self.env_id = env_id
        self.video_folder = video_folder
        self.render_every_eps = render_every_eps
        self.max_steps = max_steps
        self.use_resnet = use_resnet
        self.condition_on_mission_embed = condition_on_mission_embed
        self.fitness_fn = fitness_fn

        # initial env creation for observation space etc. if needed
        self.env = create_env(
            self.env_id,
            render_mode = 'rgb_array',
            video_folder = self.video_folder,
            render_every_eps = self.render_every_eps
        )

    def forward(self, model):

        device = module_device(model)

        seed = torch.randint(0, int(1e6), ()).item()
        state, _ = self.env.reset(seed = seed)

        if self.condition_on_mission_embed:
            mission = self.env.unwrapped.mission
            mission_embed = get_mission_embedding(mission)
            mission_embed = mission_embed.to(device)
            if mission_embed.ndim == 1:
                mission_embed = mission_embed.unsqueeze(0)

        step = 0
        cache = None
        past_action_id = None

        # correctly unwrap model to access visual_encode
        unwrapped_model = model
        if exists(self.accelerator):
            unwrapped_model = self.accelerator.unwrap_model(model)
        
        # also handle Noisable wrapper attribute blocking from x-evolution
        unwrapped_model = getattr(unwrapped_model, 'model', unwrapped_model)

        episode_rewards = []
        episode_states = []
        episode_actions = []
        episode_next_states = []
        episode_infos = []

        while step < self.max_steps:
            image = state['image']
            image_tensor = torch.from_numpy(image).float().to(device)

            if self.use_resnet:
                image_tensor = rearrange(image_tensor, 'h w c -> 1 1 h w c')
                image_tensor = unwrapped_model.visual_encode(image_tensor)
            else:
                image_tensor = rearrange(image_tensor, 'h w c -> 1 1 (h w c)')

            if torch.is_tensor(past_action_id):
                past_action_id = past_action_id.long()

            with torch.no_grad():
                logits, cache = model(
                    image_tensor,
                    past_action_id,
                    return_cache = True,
                    return_raw_action_dist = True,
                    cache = cache,
                    condition = mission_embed if self.condition_on_mission_embed else None
                )

            action = unwrapped_model.action_readout.sample(logits)
            past_action_id = action
            action_id = action.squeeze()

            next_state, reward, terminated, truncated, info = self.env.step(action_id.cpu().numpy().item())

            episode_rewards.append(reward)
            episode_states.append(state)
            episode_actions.append(action_id)
            episode_next_states.append(next_state)
            episode_infos.append(info)

            done = terminated or truncated
            if done:
                break

            state = next_state
            step += 1

        return self.fitness_fn(
            episode_rewards,
            episode_states,
            episode_actions,
            episode_next_states,
            episode_infos
        )

def main(
    env_id = 'BabyAI-BossLevel-v0',
    num_generations = 100,
    max_steps = 500,
    render_every_eps = 100,
    video_folder = './recordings_babyai_es',
    transformer_weights_path: str | None = None,
    meta_controller_weights_path: str | None = None,
    output_meta_controller_path = 'metacontroller_es_trained.pt',
    use_resnet = False,
    noise_population_size = 50,
    noise_scale = 1e-2,
    learning_rate = 1e-3,
    condition_on_mission_embed = False,
    fitness_fn = default_fitness_fn
):
    accelerator = Accelerator()

    if accelerator.is_main_process:
        rmtree(video_folder, ignore_errors = True)

    assert noise_population_size >= 2, "noise_population_size must be at least 2 for evolutionary strategies"

    # load model

    assert exists(transformer_weights_path), "Transformer weights must be provided"
    
    # lazy import to avoid unnecessary dependencies if not used
    from metacontroller.transformer_with_resnet import TransformerWithResnet as TransformerResnet
    transformer_klass = TransformerResnet if use_resnet else Transformer
    
    model = transformer_klass.init_and_load(transformer_weights_path, strict = False)
    model.eval()

    if exists(meta_controller_weights_path):
        meta_controller = MetaController.init_and_load(meta_controller_weights_path, strict = False)
        model.meta_controller = meta_controller

    assert exists(model.meta_controller), "MetaController must be present for evolution"

    # setup environment
    
    babyai_env = BabyAIEnvironment(
        env_id = env_id,
        video_folder = video_folder,
        render_every_eps = render_every_eps,
        max_steps = max_steps,
        use_resnet = use_resnet,
        condition_on_mission_embed = condition_on_mission_embed,
        fitness_fn = fitness_fn,
        accelerator = accelerator
    )

    # evolve

    model.evolve(
        num_generations = num_generations,
        environment = babyai_env,
        noise_population_size = noise_population_size,
        noise_scale = noise_scale,
        learning_rate = learning_rate,
        accelerator = accelerator
    )

    # save

    if accelerator.is_main_process:
        model.meta_controller.save(output_meta_controller_path)
        print(f'MetaController weights saved to {output_meta_controller_path}')

if __name__ == '__main__':
    fire.Fire(main)
